// Copyright (C) - lucien.povy@free.fr. Carantec 2014-2016-2020-2024
// This file must be used under the terms of the GNU GPL v2.0 licence.
// This source file is licensed as described in the file COPYING, which
// you should have received as part of this distribution.  The terms
// are also available at https://www.gnu.org/licenses/licenses.en.html 
// ====================================================================
//This function uses Scilab programs "roots" and "bezout" and gives roots of a real
//polynomial P and multiplicity of each root for initial solution.
//After that we use the theoretical method proposed by Prof.Zhonggang Zeng in MATHEMATICS OF COMPUTATION revue 
//(22/07/2004) : iterative method.
//Examples :
//s = %s; P = s**3*(s-2)**3*(s+1)*(s*s+s+1)**4*(s+2)**4*(s+6)*(s*s+1.5*s+1)**3 ; Rac = mroots(P) ;
// degree(P) = 26 ; 
//P = s**3*(s-2)**3*(s+1)*(s*s+s+1)**4*(s+2)**4*(s+6)*(s*s+1.5*s+1)**3*(s*s+.9*s..
//+.5)**4*(s*s+2)**5 (degré 44)
//Rac1 = mroots(P);  P = (s+5)**210 ; format(20);  Rac = mroots(P) !!;but P = (s+5)**220; Scilab crash (bezout or memory problem).
// ==================================================================== 
function [rm, bkerr, pjcnd, fkerr] = mroots(P, tol, iter, flag)
// rm two columns matrix,  rm = [R, mu] ; (R : roots, mu : multiplicity vector) ; bkerr : backward error ; fkerr : estimated 
//forward error of roots.
// pjcdn : pejorative condition number.
//  ****************** Inputs arguments tests.
// Number and type of input arguments. P can be a real polynomial or list. If list : then P = list(Polynomial, rm0=[approximate roots, multiplicity]).

[lhs, rhs] = argn(0);
if rhs > 4 then
  error(msprintf(gettext("%s: Wrong number of input arguments : %d to %d expected.\n"),"mroots",1,4))
  return
end
typP = typeof(P);
if and( typP <> [ "list","polynomial"] ) then
   error(msprintf(_("%s: Wrong type for input argument #%d :  a list or polynomial expected.\n"),"mroots",1))
    return
end
if typP == "polynomial" &  isreal(P) == %f  then
    error(msprintf(_("%s: Wrong type for input argument #%d : Real polynomial expected.\n"),"mroots",1))
    return
end
select rhs 
  case 1 then 
     tol = 1.e-14; iter = 10; flag = [];
  case 2 then
    if type(tol) == 10 then
      flag = tol; tol = 1.e-14; iter = 10;
    elseif  type(tol) == 1 then
        iter = 10; flag = [];
    else
      error(msprintf(_("%s: Wrong type for input argument #%d : a scalar or  string expected.\n"),"mroots",2))
      return
    end
  case 3 then
    if type(tol) == 1 & type(iter) == 1 then
      flag = [];
    elseif  type(tol) == 1 & type(iter) == 10 then
      flag = iter; iter = 10;
    else
      error(msprintf(_("%s: Wrong type for input arguments %d to %d : two scalars or a scalar and a  string expected.\n"),"mroots",2,3))
      return
    end 
  case 4 
      if type(tol) == 1 & type(iter) == 1 & type(flag) == 10 then
      else
        error(msprintf(_("%s: Wrong type for input arguments %d to %d : two scalars and string expected.\n"),"mroots",2,4))
        return
      end
end
// ************************** FIRST PART : INITIALISATION FOR MULTIPLICITY AND ROOTS.
eps = 1.e-2; // Accuracy for sum and product of the roots for a test.
select typP
  case "polynomial"  then
    coe = coeff(P);
    nz = min(find(coe < > 0)) - 1; // Number of zero roots.
    coeP = coe(nz+1:$)/coe($);
    P = poly(coeP , varn(P) ,"c"); n = degree(P); //New polynomial without zero roots and monic.
//  ***************** Main program.The extended GCD with Scilab program "bezout.sci".
    dP = derivat(P);
    [g, U] = bezout(P, dP); // Gcd of P and dP ;  U Unimodular polynomial matrix 2 by 2 such as [P dP]*U = [Gcd 0].
//  ****************** First case : Gcd degree is zero : one uses Scilab program "roots.sci" :  simple roots.
    if degree(g) == 0 then
      warning(" the polynomial has only simple roots : we use roots program.") ;
      R = roots(P);
      Rr = R(imag(R) == 0); Rii = R(imag(R) <> 0); R = [Rr; Rii]; //Real roots first.
      if nz <> 0 then
        mu = [nz; ones(R)];
        R  = [0; R]; // The roots.
      else
        mu = ones(R);
      end
      rm = [R, mu]; bkerr = []; pjcnd = []; fkerr = [];
      return
    end
//  ***************** Gcd degree is not zero.
    Ux = - U(2,2); Vx = U(1,2); // From U matrix of "bezout.sci". 
    R = roots(Ux); // Roots of P without multiplicities.
    dUx = derivat(Ux); 
    mu = horner(Vx/dUx, R); mu = round(abs(mu)); // Multiplicities of roots.
    R = R( mu <> 0); mu = mu(mu <> 0); // No roots with zero multiplicity.
// ****************** Now a veracity test.
//In some examples you have only one root (if degree of Ux is one : the degree of polynomial  Gcd is n - 1).
//Example: Wilkinson polynomials degree n=>20 : one root on gravity center ; P=poly([1:a],"s") (a>18).
// ****************** Sum and product tests of the roots.
    Sum = sum(R.*mu); Sumvrai = - coeP($-1);
    Pro = real(prod(R.^mu)); Provrai = (-1)^degree(P)*coeP(1);
    if coeP($-1) == 0 then //The sum can be zero, but not the product because we excluded zero roots.
      dSum = 0;
    else
      dSum = (Sumvrai-Sum)/Sumvrai;
    end
    dPro = (Provrai-Pro)/Provrai;
    if  and ([abs(dSum) < eps , abs(dPro) < eps]) == %F  then
      warning("Accuracy is not good : we use roots program.") ;
      R = roots(P);
      Rr = R(imag(R) == 0); Rii = R(imag(R)<>0); R = [Rr; Rii];
      if nz <> 0 then
        mu = [nz; ones(R)];
        R  = [0; R]; //The roots.
      else
        mu = ones(R);
      end
      rm = [R, mu];  bkerr = []; pjcnd = []; fkerr = [];
      return
    end
    mu = [mu(imag(R) == 0); mu(imag(R) <> 0)]; R = [R(imag(R) == 0); R(imag(R) <> 0)]; 
    if nz <> 0 then
      mu  = [nz; mu];
      R = [0; R];
    end    
    rm0 = [R, mu];
    if  part (flag,1) == "y"  then
      rm = rm0; bkerr = []; pjcnd = []; fkerr = [];
      return
    end
  case "list" then
    PP = P; rm0 = PP(2); P=PP(1);
    L = [type(P) == 2 & type(rm0) == 1 & size(rm0,"c") == 2];
    if L == %f  then
     error(msprintf(_("%s: Wrong type for input argument #%d : A list with two componants : Real polynomial and matrix expected.\n"),"mroots",1))
     return
    end
end
// ************************** SECOND PART : ITERATIVE METHOD. MULTIPLICITY STRUCTURE IS KNOWN.
// ************************** rm0 IS THE INITIAL SOLUTION FOR THE ROOTS AND MULTIPLICITY.
// Inputs : P, rm0 (initial solution computed by part one of this program), or rm0 (initial solution otherwise obtained) : rm0 = [R, mu].
// zero roots of initial solution.
s = poly(0, varn(P));
if rm0(1, 1) == 0  then  //zero roots are in first position in rm0.
  nz = rm0(1, 2); // The multiplicity of zero roots.
  R = rm0(2:$, 1); // The other roots.
  mu = rm0(2:$ , 2); // The multiplicities.
  P = (P/s^nz).num; // polynomial without zero roots.
else
  nz = 0;
  R = rm0(1:$, 1);
  mu = rm0(1:$, 2);
end
n = sum(mu); m = length(R); job = 0;
w = min(1, 1./abs(coeff(P)($-1:-1:1))); W = diag(w); w = w'; //The weights.
mur = mu(imag(R) == 0); mui = mu(imag(R) >0); muii = mu(imag(R) <> 0); mu = [mur; muii];
for k = 1:iter
// First separate real and imaginary roots and compute polynomials like Ux and g the gcd (see part 1).
    Rr = R(imag(R) == 0); Ri = R(imag(R) > 0); Rii = R(imag(R) < > 0); R = [Rr; Rii];
    U = prod(s*ones(Rr)-Rr) * prod(s^2* ones(Ri) -2*s*real(Ri) + abs(Ri).^2); // Is a real polynomial.
    Gr = prod((s*ones(Rr) - Rr).^(mur - ones(mur)));
    Gi =  prod((s^2* ones(Ri) -2*s*real(Ri) + abs(Ri).^2).^(mui - ones(mui)));
    G = Gr * Gi; // Real polynomial (like gcd).
    PS = G * U; //  PS approximation of P ;  Real polynomial.
    Dcoe = coeff(PS)($-1 : -1 : 1) - coeff(P)($-1 : -1 : 1); Dcw = w.*Dcoe'; // Also real.
// Now the jacobian, sometime matrix with imaginary part (if complex roots).
    H = s*ones(R) - R; // Elementaries polynomials.
    B = repmat(conj(H'), m, 1);
    A = B + diag(diag(-B)) + eye(B); // Matrix of Elementary polynomials, with one on diagonal. Bug in "-"diag(diag(B)).
    PR = prod(A,"c");
    M = - mu.*(G*PR);    
    coeM = coeff(M ,[n-1 : -1 : 0]);
    J = conj(coeM'); JacW = W*J; //Jacobian and weighted Jacobian.
    diffR = JacW\Dcw; //Roots correction.
    lr = length(mur);
    diffR(1 : lr) = abs(diffR(1 : lr)).*sign(real(diffR(1 : lr))); //For real roots diffR must be real.
    delta(k) = norm(diffR, 2); bcker(k) = norm(Dcw , %inf);
    if delta(k) < tol  then
      job = 1;
    end 
    if k > 1 then
      if delta(k) > delta(k-1) & bcker(k) > bcker(k-1)  then 
        if job == 1 then
          bkerr = bcker(k-1); 
          break
        end
      elseif delta(k) < delta(k-1) then
        if delta(k)^2/(delta(k-1)-delta(k)) < tol  then
          job = 1;
        end
      end
    end
    R = R - diffR;
    bkerr = bcker(k); 
end
if job ==1 then
  sv = svd(JacW); 
  pjcnd = 1/sv(m);
  fkerr = 2*bkerr*pjcnd;
// Now the zero roots
    if nz <> 0 then
        mu  = [nz; mu];
        R = [0; R];
    end
    rm = [R, mu];
end
endfunction
